/*
 * gps_gga.c
 *
 * interrupt driven $GPGGA sentence parser.
 * sorts and stores character strings for later processing
 *
 *  Created on: 30-01-2012
 *      Author: andy
 * rewritten by S. James Remington for RMC sentences 7/2015
 * This version parses GGA sentences 8/2015
 */

//*** parsing is blocked until last valid result is processed!!
//*** main program must set gps_parse_enable=1 when done

#include "gps_gga.h"


// receive interrupt, calls parse and if debugging, echos character

ISR(USART_RX_vect)
{
	char ch;
	ch = UDR0;
//  echo character for debug
//	loop_until_bit_is_set(UCSR0A, UDRE0);
//	UDR0 = ch;
	if(gps_parse_enable) gps_parse_status = gps_parse(ch);
}

//
// initialize GPS parser. If receive interrupts not used, call with uart_baud = 0
//

void gps_parse_init(int uart_baud){

    chksum = 0; //cumulative checksum
    parsemode = 0;  //initialize state machine
	gps_parse_status = 0;
	gps_parse_enable = 1;

// initialize pointer array to data fields
// array order corresponds to parse order of gpgga sentence

	gga_token[0] = &gga_time[0];
	gga_token[1] = &gga_lat[0];
	gga_token[2] = &gga_NS[0];
	gga_token[3] = &gga_long[0];
	gga_token[4] = &gga_EW[0];
	gga_token[5] = &gga_data_valid[0];
	gga_token[6] = &gga_sats[0];
	gga_token[7] = &gga_hdop[0];
	gga_token[8] = &gga_alt[0];
	gga_token[9] = &gga_altu[0];
	gga_token[10] = &gga_geoid[0];
	gga_token[11] = &gga_geoidu[0];
	gga_token[12] = &gga_dgpsage[0];
	gga_token[13] = &gga_dgpsref[0];

#define GGA_MAXTOKENINDEX 13

	if(uart_baud > 0)
	{ //set up uart for interrupt reception
  	UBRR0H = 0;
  	UBRR0L = (F_CPU / (16UL * uart_baud)) - 1;
  	UCSR0B = (1<<TXEN0) | (1<<RXEN0); /* tx/rx enable */
  	UCSR0B |= (1<<RXCIE0); //enable receive interrupt
  	}
}



// convert hex char to decimal digit

char convert_hex(char ch) {
  if (ch >= '0' && ch <= '9') return ch - '0';
  if (ch >= 'A' && ch <= 'F') return ch - 'A' + 10;
//  if (ch >= 'a' && ch <= 'f') return ch - 'a' + 10;
  return 0; //default
}
//
// decode individual NMEA lat/lon character strings in format DDDMM.MMMMM
// result is stored as long integer as degrees*10^6
// p is pointer to character string to be converted.
//
long parse_degrees(char* p) {
    char* minutes;
    char deg[]={0,0,0,0};
    double x;
    long d;

	if (strlen(p) == 0) return 0L;

    if ((minutes = strchr(p,'.')) == NULL) return 0L;  //'.' not found
    minutes -= 2; //back up pointer to include two digits of minutes+decimal fraction
    x = strtod(minutes,NULL); //make into double
    memcpy(deg,p,minutes-p); //copy degrees, works OK with zero length string
    d = strtol(deg,NULL,10)*1000000UL;  //degrees to long int
    d += (long)(x*16666.667 + 0.5); //*1.e6/60.  add in minutes

    return d;
}

// returns most recent GPS latitude and longitude
// return value = data valid flag for the most recent sentence
// return 0 = not valid
// return 1 = recent valid fix

char gps_latlon(long* lat, long* lon) {
	long temp;
	temp = parse_degrees(gga_lat);
	if (gga_NS[0] == 'S') temp = -temp;
	*lat = temp;
   	temp = parse_degrees(gga_long);
	if (gga_EW[0] == 'W') temp = -temp;
	*lon = temp;
	if (gga_data_valid[0] > '0' && gga_data_valid[0] < '6' ) return 1;
	//valid = 1 for GPS, 2 for DGPS fix, etc.
	return 0;
}

//
// return number of satellites used in fix
//

int gga_num_sats(void) {
	int j=0;
	if (gga_data_valid[0] > '0') j = 10*(gga_sats[0]-'0') + (gga_sats[1]-'0');
	if (j > 0 && j <= 99) return j; //sanity check
	return 0;
}

// GPS sentence parser. This version decodes $GPGGA sentences
// enter with incoming character
// stores character strings in respective fields
//
// returns: 0 if a gpgga sentence has not been completely parsed
//          1 if successful parse, valid data and checksum OK
//          2 if invalid data (no fix)
//          3 if checksum not OK

char gps_parse(char ch){

	static char ch4=0;

//    printf("pm%d %c\n",parsemode,ch);  //debug print

	if(ch == '$')  //check for start of sentence
        {
        char_count = 0;
        chksum = 0;
        parsemode = 1; // "$" found
        return 0;
        }

// state machine follows

	if(parsemode == 1) //now interpreting sentence type label
        {
		chksum ^= ch;  //xor the incoming char
		char_count++;  // not counting '$'
		if(char_count == 4) ch4=ch; //save the fourth char
		if(char_count == 5)
            {
            if (ch == 'A' && ch4 == 'G') //check 4th & 5th char of string
			   {
                parsemode = 3;  	// only gpgga for now
                token_index = 0; 	//initialize pointer index for data fields
                }
            else  parsemode = 0;	//not gpgga, look for next $
			}

        return 0;
		}

    if(parsemode == 3)  //in a gpgga sentence.  parsemode 2=RMC, 4-8 reserved for other types of sentences.
			{
			   if(ch == '*')  //check for checksum token
		        {
			    if (token_index > 0) *ptemp = 0; //terminate last datafield. OK if token_index > MAXTOKENINDEX
        		parsemode = 9;  //do checksum processing
        		char_count = 0; //set up to count checksum characters
        		return 0;
        		}

			chksum ^= ch;

            if(ch == ',') //end of a data field
                {
			    if (token_index > 0) *ptemp = 0; //terminate last datafield. OK if token_index > MAXTOKENINDEX
			    if (token_index <= GGA_MAXTOKENINDEX ) ptemp = gga_token[token_index]; //point to next data field
			    token_index++; //point to next data field
				// printf("token_index %d\n",token_index);
			    return 0;
			    }

            if (token_index <= GGA_MAXTOKENINDEX+1)
			    //still storing data (note, number of data fields = maxindex+1)
                {
                *ptemp = ch;  //copy character into data field
                ptemp++;
				}

            return 0;
            }

	if (parsemode == 9) //now processing checksum
                {
                gga_chksum[char_count] = ch;
                char_count++;
                if (char_count == 2)
                    {
                    char test = 16*convert_hex(gga_chksum[0]) + convert_hex(gga_chksum[1]);
					// printf("checksum 0x%02X  GPS 0x%02X\n",chksum,test);

                    parsemode = 0; //done reading checksum chars

                    if( (chksum == test) && (gga_data_valid[0] > '0') )
						{
					    gps_parse_enable = 0; // ***** done, stop parsing *****
						return 1; //checksum OK and valid fix
                    	}
					if (chksum == test) return 2; //chksum OK, no position fix
					return 3; //bad checksum
                    }
                }
    return 0;
}
/*
   Display gps strings. Will display garbage if array indices are overstepped (useful for debugging).
*/
/*
void gga_disp_data(void){
	printf("\ntime = %s",gga_time);
	printf("\nvalid flag = %s",gga_data_valid);
	printf("\nLat = %s %s",gga_NS,gga_lat);
	printf("\nLong = %s %s",gga_EW,gga_long);
	printf("\nalt = %s ",gga_alt);
	return;

}
*/

// find the bearing and distance in meters from point 1 to 2,
// using the equirectangular approximation
// lat and lon are degrees*1.0e6, 10 cm precision
// for higher accuracy, radius to geodetic latitude can be calculated (now commented out)

double course_to(long lat1, long lon1, long lat2, long lon2, double* distance) {

	double dlam,dphi,radius;
	//double b=6378000.0,a=6357000.0; //equatorial and polar radii of Earth

	dphi = DEG2RAD*(lat1+lat2)*0.5e-6; //average latitude in radians
	double cphi=cos(dphi);
	//double sphi=sin(dphi);

	dphi = DEG2RAD*(lat2-lat1)*1.0e-6; //differences in degrees
	dlam = DEG2RAD*(lon2-lon1)*1.0e-6;

	dlam *= cphi;  //correct for latitude

	double bearing=RAD2DEG*atan2(dlam,dphi);
	if(bearing<0) bearing=bearing+360.0;

/* optional radius to geodetic (standard) latitude
	double a2=a*a;
	double b2=b*b;
	double s2=sphi*sphi;
	double c2=cphi*cphi;
	radius = sqrt((a2*a2*c2 + b2*b2*s2)/(a2*c2 + b2*s2));
*/

	radius = 6371000.0;
	*distance = radius * sqrt(dphi*dphi + dlam*dlam);
	return bearing;
}

//
// find latitude and longitude (lat2,lon2) of a point,
// given bearing (degrees) and distance (meters), lat1, lon1
//
// **great circle route on spherical Earth assumed**
//
// input and output (lat,lon) are long integers in degrees*1e6
// on ATMegas, this is less accurate due to single precision floats
/*
void bearing_shoot(float bearing, float distance, long lat1, long lon1, long *lat2, long *lon2)
{

	float phi1,lam1,d,phi2;
	float R=6371000.0;  //mean Earth radius in meters
	float theta = bearing*DEG2RAD;

	d = distance/R;  //arc length

	phi1 = DEG2RAD*lat1*1.0e-6;  //convert to radians
	lam1 = DEG2RAD*lon1*1.0e-6;
	phi2 = asin(sin(phi1)*cos(d) + cos(phi1)*sin(d)*cos(theta));

	*lat2 = (long) 1.0e6*phi2*RAD2DEG;
	*lon2 = (long) 1.0e6*(lam1 + atan2(sin(theta)*sin(d)*cos(phi1), cos(d)-sin(phi1)*sin(phi2)))*RAD2DEG;
}
*/

// this version of bearing_shoot uses the equirectangular approximation.
// GPS coordinates assumed to be long integers, degrees*10^6 (10 cm precision)
// tests show it to be more accurate for short distances than the great circle
// version above, but consumes far less program memory

void bearing_shoot (float bearing, float distance, long lat1, long lon1, long *lat2, long *lon2)
{
  float d,phi1;
  float R=6371000.0;  //mean Earth radius in meters
  long dlat, dlon;

  float theta = bearing*DEG2RAD;
  d = RAD2DEG*distance/R;  //arc length in degrees

  phi1 = DEG2RAD*lat1*1.0e-6;  //convert lat to radians
  dlat = d*cos(theta)*1.0e6; //latitude change in degrees*10^6
  dlon = d*sin(theta)*1.0e6/cos(phi1); //longitude change, corrected for latitude

    *lat2 = lat1 + dlat;
    *lon2 = lon1 + dlon;
}
